import os
from functools import partial

import fiona
import numpy as np
import pandas as pd
import pyproj
from scipy import stats
from shapely.geometry import LineString, Point
from shapely.geometry import shape
from shapely.ops import transform


def shapes_from_shp(shp_file):
    """
    Parses a shape file and returns a list of shapely shapes
    :param shp_file:
    :return:
    """
    shapes = []
    ids = []
    properties = []
    for feat in fiona.open(shp_file, 'r'):
        shapes.append(shape(feat['geometry']))
        ids.append(feat['id'])
        properties.append(feat['properties'])
    return shapes, ids, properties


def convert_coord_systems(g1, in_coord_system='EPSG:4326', out_coord_system='EPSG:28356'):
    """
    Converts coordinates from one coordinates system to another. Needed because shapefiles are usually defined in
    lat/lon but should be converted to GDA to calculated distances.
    https://gis.stackexchange.com/a/127432
    :param in_coord_system: Default is lat/lon WGS84
    :param out_coord_system: Default is GDA56 for NSW coastline
    :return:
    """
    project = partial(
        pyproj.transform,
        pyproj.Proj(init=in_coord_system),  # source coordinate system
        pyproj.Proj(init=out_coord_system))  # destination coordinate system

    g2 = transform(project, g1)  # apply projection
    return g2


def get_slope(x, z, top_elevation, btm_elevation, method='end_points'):
    """
    Returns a slope from a bed profile
    :param x: List of x bed profile coordinates
    :param z: List of z bed profile coordinates
    :param top_elevation: Top elevation of where to take the slope
    :param btm_elevation: Bottom elevation of where to take the slope
    :param method: Method used to calculate slope (end_points or least_squares)
    :return:
    """

    profile_line = LineString([[x, z] for x, z in zip(x, z)])

    end_points = {
        'top': {
            'elevation': top_elevation,
        },
        'btm': {
            'elevation': btm_elevation,
        }}

    # If there are multiple intersections, take most seaward of landward point?
    multiple_points = 'landward'

    for end_type in end_points.keys():
        elevation = end_points[end_type]['elevation']
        elevation_line = LineString([[min(x), elevation], [max(x), elevation]])
        intersection_points = profile_line.intersection(elevation_line)

        if intersection_points.is_empty:
            return None

        if multiple_points == 'landward':
            intersection_point = list(intersection_points)[0]
        elif multiple_points == 'seaward':
            intersection_point = list(intersection_points)[-1]

        end_points[end_type]['x'] = intersection_point.xy[0][0]
        end_points[end_type]['z'] = intersection_point.xy[1][0]

    if method == 'end_points':
        x_top = end_points['top']['x']
        x_btm = end_points['btm']['x']
        z_top = end_points['top']['z']
        z_btm = end_points['btm']['z']
        return -(z_top - z_btm) / (x_top - x_btm)

    elif method == 'least_squares':
        profile_mask = [True if end_points['top']['x'] < pts < end_points['btm']['x'] else False for pts in x]
        profile_x = np.array(x)[profile_mask].tolist()
        profile_z = np.array(z)[profile_mask].tolist()
        slope, _, _, _, _ = stats.linregress(profile_x, profile_z)
        return -slope


def distance_to_intersection(lat, lon, landward_orientation, beach, line_strings, line_properties):
    """
    Returns the distance at whjch a line drawn from a lat/lon at an orientation intersects a line stinrg
    :param lat:
    :param lon:
    :param landward_orientation: Angle, anticlockwise positive from east in degrees, towards the land
    towards the
    land.
    :param line_string:
    :return:
    """
    start_point = Point(lon, lat)
    start_point = convert_coord_systems(start_point)

    distance = 1000  # m look up to 1000m for an intersection
    landward_point = Point(start_point.coords.xy[0] + distance * np.cos(np.deg2rad(landward_orientation)),
                           start_point.coords.xy[1] + distance * np.sin(np.deg2rad(landward_orientation)))
    landward_line = LineString([start_point, landward_point])
    seaward_point = Point(start_point.coords.xy[0] - distance * np.cos(np.deg2rad(landward_orientation)),
                          start_point.coords.xy[1] - distance * np.sin(np.deg2rad(landward_orientation)))
    seaward_line = LineString([start_point, seaward_point])

    # Look at relevant line_strings which have the same beach property in order to reduce computation time
    line_strings = [s for s, p in zip(line_strings, line_properties) if p['beach'] == beach]

    # Check whether profile_line intersects with any lines in line_string. If intersection point is landwards,
    # consider this negative, otherwise seawards is positive.
    for line_string in line_strings:
        land_intersect_points = landward_line.intersection(line_string)
        if not land_intersect_points.is_empty:
            return -land_intersect_points.distance(start_point)

        sea_intersect_points = seaward_line.intersection(line_string)
        if not sea_intersect_points.is_empty:
            return sea_intersect_points.distance(start_point)

    # If no intersections are found, return nothing.
    return None

# todo
def get_sites_dune_crest_toe():
    """
    Reads dune crest and toe files and creates a pandas dataframe with crest/toe locations at each site_
    :return:
    """

    # Get site information
    data_folder = './data/interim'
    df_sites = pd.read_csv(os.path.join(data_folder, 'sites.csv'), index_col=[0])

    # Import our dune crest and toes
    for f in ['./data/raw/profile_features/dune_crests.shp']:
        shapes, _ = shapes_from_shp(f)
        shapes = [convert_coord_systems(x) for x in shapes]

    # Iterate through each site
